@@ -3,16 +3,15 @@
 .. moduleauthor:: Chris Bowman <chris.bowman.physics@gmail.com>
 """
 
-from numpy import abs, exp, log, dot, sqrt, argmin, diagonal, ndarray, arange
-from numpy import zeros, ones, array, where, pi, diag, eye, maximum, minimum
+from numpy import abs, exp, log, dot, sqrt, diagonal, ndarray, arange
+from numpy import zeros, ones, array, pi, diag, eye, maximum, minimum
 from numpy import sum as npsum
 from numpy.random import random
 from scipy.special import erf, erfcx
-from numpy.linalg import inv, slogdet, solve, cholesky
+from numpy.linalg import cholesky
 from scipy.linalg import solve_triangular
-from scipy.optimize import minimize, differential_evolution, fmin_l_bfgs_b
+from scipy.optimize import differential_evolution, fmin_l_bfgs_b
 from multiprocessing import Pool
-from itertools import product
 from warnings import warn
 from copy import copy
 from inspect import isclass
@@ -275,14 +274,20 @@ class GpRegressor(object):
         number of dimensions). Alternatively, a list of array-like objects can be
         given, which will be converted to a ``ndarray`` internally.
 
-    :param y: The y-data values as a list or array of floats.
+    :param y: \
+        The y-data values as a 1D array.
 
     :param y_err: \
-        The error on the y-data values supplied as a list or array of floats.
+        The error on the y-data values supplied as a 1D array.
         This technique explicitly assumes that errors are Gaussian, so the supplied
         error values represent normal distribution standard deviations. If this
         argument is not specified the errors are taken to be small but non-zero.
 
+    :param y_cov: \
+        A covariance matrix representing the uncertainties on the y-data values.
+        This is an alternative to the 'y_err' keyword argument, allowing the
+        y-data covariance matrix to be specified directly.
+
     :param hyperpars: \
         An array specifying the hyper-parameter values to be used by the
         covariance function class, which by default is ``SquaredExponential``.
@@ -309,8 +314,8 @@ class GpRegressor(object):
         Sets the number of processes used in optimizing the hyper-parameter values.
         Multiple processes are only used when the optimizer keyword is set to "bfgs".
     """
-    def __init__(self, x, y, y_err = None, hyperpars = None, kernel = SquaredExponential, mean = ConstantMean,
-                 cross_val = False, optimizer = 'bfgs', n_processes = 1):
+    def __init__(self, x, y, y_err=None, y_cov=None, hyperpars=None, kernel=SquaredExponential, mean=ConstantMean,
+                 cross_val=False, optimizer='bfgs', n_processes=1):
 
         # store the data
         self.x = array(x)
@@ -328,32 +333,20 @@ class GpRegressor(object):
             raise ValueError('The given number of x-data points does not match the number of y-data values')
 
         # build data errors covariance matrix
-        self.sig = zeros([self.N_points, self.N_points])
-        if y_err is not None:
-            if len(y) == len(y_err):
-                for i in range(len(self.y)):
-                    self.sig[i,i] = y_err[i]**2
-            else:
-                raise ValueError('y_err must be the same length as y')
+        self.sig = self.check_error_data(y_err, y_cov)
 
         # create an instance of the covariance function if only the class was passed
-        if isclass(kernel):
-            self.cov = kernel()
-        else:
-            self.cov = kernel
+        self.cov = kernel() if isclass(kernel) else kernel
 
         # create an instance of the mean function if only the class was passed
-        if isclass(mean):
-            self.mean = mean()
-        else:
-            self.mean = mean
+        self.mean = mean() if isclass(mean) else mean
 
         # pass the data to the mean and covariance functions for pre-calculations
         self.cov.pass_data(self.x, self.y)
         self.mean.pass_data(self.x, self.y)
         # collect the bounds on all the hyper-parameters
         self.hp_bounds = copy(self.mean.bounds)
-        self.hp_bounds.extend( copy(self.cov.bounds) )
+        self.hp_bounds.extend(copy(self.cov.bounds))
         # build slices to address the different parameter sets
         self.n_hyperpars = len(self.hp_bounds)
         self.mean_slice = slice(0, self.mean.n_params)
@@ -437,7 +430,81 @@ class GpRegressor(object):
         self.K_xx = self.cov.build_covariance(self.cov_hyperpars) + self.sig
         self.mu = self.mean.build_mean(self.mean_hyperpars)
         self.L = cholesky(self.K_xx)
-        self.alpha = solve_triangular(self.L.T, solve_triangular(self.L, self.y-self.mu, lower = True))
+        self.alpha = solve_triangular(self.L.T, solve_triangular(self.L, self.y-self.mu, lower=True))
+
+    def check_error_data(self, y_err, y_cov):
+        if y_cov is not None:
+            # if y_cov is given as a list or tuple, attempt conversion to an array
+            if any([type(y_cov) is t for t in [list, tuple]]):
+                y_err = array(y_cov).squeeze()
+            elif type(y_cov) is not ndarray:  # else if it isn't already an array raise an error
+                raise TypeError(
+                    """
+                    [ GpRegressor error ]
+                    The 'y_cov' keyword argument should be given as a numpy array:
+                    Expected type {} but type {} was given.
+                    """.format(ndarray, type(y_err))
+                )
+
+            # now check to make sure the given error array is a valid size
+            if y_cov.shape != (self.N_points, self.N_points):
+                raise ValueError(
+                    """
+                    [ GpRegressor error ]
+                    The 'y_cov' keyword argument was passed an array with an incorrect shape.
+                    'y_cov' must be a 2D array of shape (N,N), where 'N' is the number of given 
+                    y-data values.
+                    """
+                )
+
+            # check to make sure the given matrix is symmetric
+            if not (y_cov == y_cov.T).all():
+                raise ValueError(
+                    """
+                    [ GpRegressor error ]
+                    The covariance matrix passed to the 'y_cov' keyword argument is not symmetric.
+                    """
+                )
+            
+            # raise a warning if both keywords have been specified
+            if y_err is not None:
+                warn(
+                    """
+                    [ GpRegressor warning ]
+                    Only one of the 'y_err' and 'y_cov' keyword arguments should be specified.
+                    Only the input to 'y_cov' will be used, the input to 'y_err' will be ignored.
+                    """
+                )
+
+            return y_cov
+
+        elif y_err is not None:
+            # if y_err is given as a list or tuple, attempt conversion to an array
+            if any([type(y_err) is t for t in [list, tuple]]):
+                y_err = array(y_err).squeeze()
+            elif type(y_err) is not ndarray:  # else if it isn't already an array raise an error
+                raise TypeError(
+                    """
+                    [ GpRegressor error ]
+                    The 'y_err' keyword argument should be given as a numpy array:
+                    Expected type {} but type {} was given.
+                    """.format(ndarray, type(y_err))
+                )
+
+            # now check to make sure the given error array is a valid size
+            if y_err.shape != (self.N_points,):
+                raise ValueError(
+                    """
+                    [ GpRegressor error ]
+                    The 'y_err' keyword argument was passed an array with an incorrect shape.
+                    'y_err' must be a 1D array of length N, where 'N' is the number of given 
+                    y-data values.
+                    """
+                )
+
+            return diag(y_err**2)
+        else:
+            return zeros([self.N_points, self.N_points])
 
     def process_points(self, points):
         if type(points) is ndarray:
@@ -447,7 +514,7 @@ class GpRegressor(object):
 
         m = len(x.shape)
         if self.N_dimensions == 1:
-            if m == 0: # the case when given a float
+            if m == 0:  # the case when given a float
                 x = x.reshape([1,1])
             elif m == 1:
                 x = x.reshape([x.shape[0],1])
@@ -732,196 +799,6 @@ class MarginalisedGpRegressor(object):
 
 
 
-class GpInverter(object):
-    """
-    Solves linear inverse problems of the form y = Gb, using a Gaussian-process
-    prior which imposes spatial regularity on the solution.
-
-    The solution vector 'b' must describe the value of a quantity everywhere
-    on a grid, as the GP prior imposes covariance between these grid-points
-    based on the 'distance' between them. The grid need not be a spatial one,
-    only one over which regularity is desired, e.g. time, wavelength ect.
-
-    > arguments
-
-        x 	- array of position values/vectors for the model parameters
-
-        y 	- array of data values
-
-        cov	- covariance matrix for the data
-
-        G	- the linearisation matrix
-    """
-    def __init__(self, x, y, cov, G, scale_length = None, mean = None, amplitude = None, selector = 'evidence'):
-        self.x = x  # spatial location of the parameters, *not* the y data
-        self.y = y  # data values
-        self.S_y = cov  # data covariance matrix
-        self.G = G  # geometry matrix
-
-        self.selector = selector
-        self.hyperpar_settings = (amplitude, scale_length, mean)
-
-        # check inputs for compatability
-        self.parse_inputs()
-
-        self.I = ones([G.shape[1],1])
-        self.f = dot( self.G, self.I )
-        self.iS_y = inv(self.S_y)
-
-        # generate square-distance matrix from self.x
-        if hasattr(self.x[0], '__iter__'): # multi-dimensional case
-            self.D = [ [ self.dist(i,j) for j in self.x] for i in self.x ]
-        else: # 1D case
-            self.D = [ [ (i-j)**2 for j in self.x] for i in self.x ]
-        self.D = -0.5*array(self.D)
-
-        self.A, self.L, self.mu_val = self.optimize_hyperparameters()
-
-        # now we have determined the hyperparameters, generate the prior
-        # mean and covariance matrices
-        self.mu_p = self.mu_val * ones([len(x), 1])
-        self.S_p = (self.A**2)*exp(self.D/(self.L**2))
-
-        # we may now also generate the posterior mean and covariance.
-        # To improve the numerical stability of calculating the posterior
-        # covariance, we use the woodbury matrix identity:
-        K = dot(self.G, self.S_p)
-
-        V = self.S_y + dot(K, self.G.T)
-        iVK = solve(V,K)
-        self.S_b = self.S_p - dot( K.T, iVK )
-
-        # posterior mean involves no further inversions so is stable
-        self.mu_b = self.mu_p + dot( self.S_b, dot( self.G.T, dot( self.iS_y, (self.y - self.mu_val*self.f) ) ) )
-
-    def parse_inputs(self):
-        # first check input types
-        if type(self.y) is not ndarray: self.y = array(self.y)
-        if type(self.S_y) is not ndarray: self.S_y = array(self.S_y)
-        if type(self.G) is not ndarray: self.G = array(self.G)
-
-        # now check shapes / sizes are compatible
-        if len(self.y.shape) is not 2: self.y = self.y.reshape([self.y.size,1])
-        if self.S_y.shape[0] != self.S_y.shape[0]:
-            raise ValueError('Data covariance matrix must be square')
-        if self.S_y.shape[0] != self.y.shape[0]:
-            raise ValueError('Dimensions of the data covariance matrix must equal the number of data points')
-        if (self.G.shape[0] != self.y.shape[0]) or  (self.G.shape[1] != len(self.x)):
-            raise ValueError('The operator matrix must have dimensions [# data points, # spatial points]')
-
-    def dist(self, a, b):
-        return sum( (i-j)**2 for i, j in zip(a, b) )
-
-    def log_ev(self, h):
-        # extract hyperparameters
-        A, L, mu_p = [exp(v) for v in h]
-        # first make the prior covariance
-        S_p = (A**2)*exp(self.D/(L**2))
-        # now the marginal likelihood covariance
-        S_m = dot( self.G, dot(S_p, self.G.T) ) + self.S_y
-        # and the marginal likelihood mean
-        mu_m = mu_p * self.f
-        # now calculate negative log marginal likelihood
-        u = self.y - mu_m
-        iSu = solve(S_m, u)
-        L = dot( u.T, iSu ) + slogdet(S_m)[1]
-        return L[0][0]
-
-    def nn_maximum_likelihood(self, h):
-        A, L, mu_p = [exp(v) for v in h]
-
-        S_p = (A**2)*exp(self.D/(L**2))
-
-        K = dot(self.G, S_p)
-        V = self.S_y + dot(K, self.G.T)
-        iVK = solve(V,K)
-        S_b = S_p - dot( K.T, iVK )
-
-        # posterior mean involves no further inversions so is stable
-        mu_b = mu_p + dot( S_b, dot( self.G.T, dot( self.iS_y, (self.y - mu_p*self.f) ) ) )
-        mu_b[where(mu_b < 0)] = 0.
-        # find the residual
-        res = self.y - self.G.dot(mu_b)
-        LL = dot(res.T, self.iS_y.dot(res))
-        return LL[0,0]
-
-    def optimize_hyperparameters(self):
-        # choose the selection criterion for the hyperparameters
-        if self.selector is 'evidence':
-            criterion = self.log_ev
-        elif self.selector is 'NNML':
-            criterion = self.nn_maximum_likelihood
-        else:
-            raise ValueError('The selector keyword must be given as either `evidence` or `NNML`')
-
-        # Choose the correct inputs for the criterion based on which
-        # hyperparameters have been given fixed values
-        code = tuple([ x is None for x in self.hyperpar_settings ])
-        log_vals = []
-        for x in self.hyperpar_settings:
-            if x is None:
-                log_vals.append(None)
-            else:
-                log_vals.append(log(x))
-
-        selection_functions = {
-            (1,1,1) : lambda x : criterion(x),
-            (1,1,0) : lambda x : criterion([x[0],x[1],log_vals[2]]),
-            (1,0,1) : lambda x : criterion([x[0],log_vals[1],x[1]]),
-            (0,1,1) : lambda x : criterion([log_vals[0],x[0],x[1]]),
-            (1,0,0) : lambda x : criterion([x[0],log_vals[1],log_vals[2]]),
-            (0,1,0) : lambda x : criterion([log_vals[0],x[0],log_vals[2]]),
-            (0,0,1) : lambda x : criterion([log_vals[0],log_vals[1],x[0]]),
-            (0,0,0) : None
-        }
-
-        minfunc = selection_functions[code]
-
-        # if all the hyperparameters have been fixed, just return the fixed values
-        if minfunc is None: return self.hyperpar_settings
-
-
-        # make some guesses for the hyperparameters
-        A_guess  = [-6,-4,-2, 0]
-        L_guess  = [-6,-5,-4,-3,-2] # NOTE - should be data-determined in future
-        mu_guess = [-8,-6,-4,-2, 0]
-
-        # build a list of initial guesses again depending on what parameters are fixed
-        guess_components = []
-        if code[0]: guess_components.append(A_guess)
-        if code[1]: guess_components.append(L_guess)
-        if code[2]: guess_components.append(mu_guess)
-        guesses = [ g for g in product(*guess_components) ]
-
-        # sort the guesses by best score
-        guesses = sorted(guesses, key = minfunc)
-
-        LML_list   = []
-        theta_list = []
-
-        for g in guesses[:3]: # minimize the LML for the best guesses
-            min_obj = minimize( minfunc, g, method = 'L-BFGS-B' )
-            LML_list.append( min_obj['fun'] )
-            theta_list.append( min_obj['x'] )
-
-        # pick the solution the best score
-        opt_params = theta_list[ argmin(array(LML_list)) ]
-        paras = []
-        k = 0
-        for i in range(3):
-            if code[i]:
-                paras.append(opt_params[k])
-                k += 1
-            else:
-                paras.append(log_vals[i])
-
-        return [exp(v) for v in paras]
-
-
-
-
-
-
 class ExpectedImprovement(object):
     r"""
     ``ExpectedImprovement`` is an acquisition-function class which can be passed to
